/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

// derived from Houndini sample: SOP_Surface

/** @file */

#ifndef __houdini_HoudiniSOP_h__
#define __houdini_HoudiniSOP_h__

namespace fe
{

/**************************************************************************//**
	@brief Houdini SOP node

	@ingroup houdini
*//***************************************************************************/
class HoudiniSOP: public SOP_Node
{
	public:
								HoudiniSOP(OP_Network *pNetwork,
										const char *name,OP_Operator *pOp);
virtual							~HoudiniSOP(void);

static	OP_Node*				create(OP_Network *pNetwork,const char *name,
										OP_Operator *pOp)
								{	return new HoudiniSOP(pNetwork,name,pOp); }

static	int						discoverOperators(OP_OperatorTable *pTable);

								///	@internal
static	int						reloadCallback(void *pData,int index,float time,
										const PRM_Template* pTemplate);

								///	@internal
		void					brush(GU_Detail* pGdp,WindowEvent& a_rEvent,
										SpatialTransform& a_rCameraTransform,
										SpatialVector& a_rRayOrigin,
										SpatialVector& a_rRayDirection);
	protected:
								//* as SOP_Node
virtual OP_ERROR				cookMySop(OP_Context &context);
virtual	OP_ERROR				cookMyGuide1(OP_Context &context);

								//* as OP_Parameters
virtual	unsigned int			disableParms(void);

	private:
		void					updateCatalog(Real time,BWORD a_doInput);

		HoudiniContext					m_houdiniContext;
		HoudiniBrush*					m_pHoudiniBrush;

		sp<OperatorSurfaceI>			m_spOperatorSurfaceI;
		sp<SignalerI>					m_spSignalerI;
		sp<Layout>						m_spCookLayout;
		sp<Layout>						m_spBrushLayout;
		sp<DrawI>						m_spDrawOutput;
		sp<DrawI>						m_spDrawGuideImmediate;
		sp<DrawI>						m_spDrawGuideCached;
		sp<DrawI>						m_spDrawBrush;

		Record							m_cookSignal;
		Record							m_brushSignal;

		Accessor< sp<Component> >		m_aSurfaceOutput;
		Accessor< sp<Component> >		m_aDrawOutput;
		Accessor< sp<Component> >		m_aDrawGuide;
		Accessor< sp<Component> >		m_aDrawBrush;
		Accessor<Record>				m_aWindowEvent;
		Accessor<SpatialTransform>		m_aCameraTransform;
		Accessor<SpatialVector>			m_aRayOrigin;
		Accessor<SpatialVector>			m_aRayDirection;

		std::vector< sp<Component> >	m_surfaceInputArray;
		sp<Component>					m_spSurfaceOutput;
		String							m_nameOutput;
		String							m_lastName;
		Real							m_lastTime;

		OP_Operator*					m_pOpOperator;
		GU_Detail*						m_pGdp;
};

} // namespace

#endif /* __houdini_HoudiniSOP_h__ */
