/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __maya_MayaNode_h__
#define __maya_MayaNode_h__

namespace fe
{

/**************************************************************************//**
    @brief Maya node

	@ingroup maya
*//***************************************************************************/
class MayaNode: public MPxNode
{
	private:
	class Parameter
	{
		public:
			MObject						m_object;
	};
	class Registration
	{
		public:
			U32							m_uniqueId;
			String						m_implementation;
			std::map<String,Parameter>	m_parameterMap;
			std::vector<String>			m_inputChannelArray;
			std::vector<String>			m_outputChannelArray;
			BWORD						m_autoOutput;
	};

	public:
								MayaNode(void);
virtual							~MayaNode(void);

static	void*					create(void);

static	void					preRegister(U32 a_uniqueId,String a_name,
										String a_implementation);
static	MStatus					initialize(void);
static	void					postRegister(void);

virtual	MStatus					compute(const MPlug& rPlug,MDataBlock& rData);

	private:
static	void					populateOperator(String a_name);
		void					updateCatalog(Real time,BWORD a_doInput,
										MDataBlock& rData);

static	I32								ms_registerLock;	//* TODO threadsafe
static	String							ms_registerName;

static	MObject							ms_compoundBehavior;

static	std::map<String,Registration>	ms_mayaRegistry;

		MayaContext						m_mayaContext;
		sp<OperatorSurfaceI>			m_spOperatorSurfaceI;
		sp<SignalerI>					m_spSignalerI;
		sp<Layout>						m_spLayout;
		sp<DrawI>						m_spDrawI;
		Accessor< sp<Component> >		m_aSurfaceOutput;
		Accessor< sp<Component> >		m_aDrawI;
		std::vector< sp<Component> >	m_surfaceInputArray;
		sp<Component>					m_spSurfaceOutput;

		String							m_nameOutput;
		String							m_name;
		String							m_implementation;
		String							m_lastName;
		Real							m_lastTime;
};

} // namespace

#endif /* __maya_MayaNode_h__ */
