/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __operator_OperatorSurfaceCommon_h__
#define __operator_OperatorSurfaceCommon_h__

namespace fe
{

/**************************************************************************//**
    @brief Partial Generic SurfaceI Implemention

	@ingroup operator
*//***************************************************************************/
class FE_DL_EXPORT OperatorSurfaceCommon:
	public Catalog,
	virtual public OperatorSurfaceI
{
	public:
				//* duplicates from SurfaceAccessibleI
		enum	Element
				{
					e_point=SurfaceAccessibleI::e_point,
					e_pointGroup=SurfaceAccessibleI::e_pointGroup,
					e_vertex=SurfaceAccessibleI::e_vertex,
					e_primitive=SurfaceAccessibleI::e_primitive,
					e_primitiveGroup=SurfaceAccessibleI::e_primitiveGroup,
					e_detail=SurfaceAccessibleI::e_detail
				};
		enum	Attribute
				{
					e_generic=SurfaceAccessibleI::e_generic,
					e_position=SurfaceAccessibleI::e_position,
					e_normal=SurfaceAccessibleI::e_normal,
					e_vertices=SurfaceAccessibleI::e_vertices
				};
		enum	Message
				{
					e_quiet=SurfaceAccessibleI::e_quiet,
					e_warning=SurfaceAccessibleI::e_warning,
					e_error=SurfaceAccessibleI::e_error
				};
		enum	Creation
				{
					e_refuseMissing=SurfaceAccessibleI::e_refuseMissing,
					e_createMissing=SurfaceAccessibleI::e_createMissing
				};

							//* As BindI
virtual	void				bind(sp<SignalerI> spSignalerI,sp<Layout> spLayout)
							{
								m_aSurfaceOutput.initialize(spLayout->scope(),
										"SurfaceOutput");
								m_aDrawOutput.initialize(spLayout->scope(),
										"DrawI");
								m_aDrawGuide.initialize(spLayout->scope(),
										"DrawGuide");
								m_aDrawBrush.initialize(spLayout->scope(),
										"DrawBrush");
								m_aWindowEvent.initialize(spLayout->scope(),
										"WindowEvent");
								m_aCameraTransform.initialize(spLayout->scope(),
										"CameraTransform");
								m_aRayOrigin.initialize(spLayout->scope(),
										"RayOrigin");
								m_aRayDirection.initialize(spLayout->scope(),
										"RayDirection");
							}

							//* As OperatorSurfaceCommon
		sp<Component>		surfaceOutput(Record& a_rSignal)
							{
								sp<Component>* pspComponent=
										m_aSurfaceOutput.queryAttribute(
										a_rSignal);
								if(pspComponent)
								{
									return *pspComponent;
								}
								return sp<Component>(NULL);
							}
		sp<Component>		outputDraw(Record& a_rSignal)
							{
								sp<Component>* pspComponent=
										m_aDrawOutput.queryAttribute(a_rSignal);
								if(pspComponent)
								{
									return *pspComponent;
								}
								return sp<Component>(NULL);
							}
		sp<Component>		guideDraw(Record& a_rSignal)
							{
								sp<Component>* pspComponent=
										m_aDrawGuide.queryAttribute(a_rSignal);
								if(pspComponent)
								{
									return *pspComponent;
								}
								return sp<Component>(NULL);
							}
		sp<Component>		brushDraw(Record& a_rSignal)
							{
								sp<Component>* pspComponent=
										m_aDrawBrush.queryAttribute(a_rSignal);
								if(pspComponent)
								{
									return *pspComponent;
								}
								return sp<Component>(NULL);
							}
		Record				windowEvent(Record& a_rSignal)
							{
								Record* pRecord=
										m_aWindowEvent.queryAttribute(
										a_rSignal);
								if(pRecord)
								{
									return *pRecord;
								}
								return Record();
							}
		SpatialTransform	cameraTransform(Record& a_rSignal)
							{
								SpatialTransform* pSpatialTransform=
										m_aCameraTransform.queryAttribute(
										a_rSignal);
								if(pSpatialTransform)
								{
									return *pSpatialTransform;
								}
								return SpatialTransform::identity();
							}
		SpatialVector		rayOrigin(Record& a_rSignal)
							{
								SpatialVector* pSpatialVector=
										m_aRayOrigin.queryAttribute(a_rSignal);
								if(pSpatialVector)
								{
									return *pSpatialVector;
								}
								return SpatialVector(0.0,0.0,0.0);
							}
		SpatialVector		rayDirection(Record& a_rSignal)
							{
								SpatialVector* pSpatialVector=
										m_aRayDirection.queryAttribute(
										a_rSignal);
								if(pSpatialVector)
								{
									return *pSpatialVector;
								}
								return SpatialVector(0.0,0.0,0.0);
							}

							//* boilerplate reduction methods
		BWORD				accessDraw(sp<DrawI>& a_rspDraw,Record& a_rSignal,
									Message a_message=e_error)
							{
								a_rspDraw=outputDraw(a_rSignal);
								if(a_rspDraw.isValid())
								{
									return TRUE;
								}
								reportDraw("Output",a_message);
								return FALSE;
							}
		BWORD				accessGuide(sp<DrawI>& a_rspDraw,Record& a_rSignal,
									Message a_message=e_error)
							{
								a_rspDraw=guideDraw(a_rSignal);
								if(a_rspDraw.isValid())
								{
									return TRUE;
								}
								reportDraw("Guide",a_message);
								return FALSE;
							}
		BWORD				accessBrush(sp<DrawI>& a_rspDraw,Record& a_rSignal,
									Message a_message=e_error)
							{
								a_rspDraw=brushDraw(a_rSignal);
								if(a_rspDraw.isValid())
								{
									return TRUE;
								}
								reportDraw("Brush",a_message);
								return FALSE;
							}
		BWORD				accessOutput(
									sp<SurfaceAccessibleI>& a_rspAccessible,
									Record& a_rSignal,Message a_message=e_error)
							{
								a_rspAccessible=surfaceOutput(a_rSignal);
								m_key="<output>";
								if(a_rspAccessible.isValid())
								{
									return TRUE;
								}
								reportAccessible(a_message);
								return FALSE;
							}
		BWORD				accessOutput(sp<SurfaceAccessorI>& a_rspAccessor,
									Record& a_rSignal,
									Element a_element,Attribute a_attribute,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								sp<SurfaceAccessibleI> spAccessible;
								if(!accessOutput(spAccessible,
										a_rSignal,a_message))
								{
									return FALSE;
								}

								m_key="<output>";
								return access(a_rspAccessor,spAccessible,
										a_element,a_attribute,
										a_message,a_create);
							}
		BWORD				accessOutput(sp<SurfaceAccessorI>& a_rspAccessor,
									Record& a_rSignal,
									Element a_element,String a_name,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								sp<SurfaceAccessibleI> spAccessible;
								if(!accessOutput(spAccessible,
										a_rSignal,a_message))
								{
									return FALSE;
								}

								m_key="<output>";
								return access(a_rspAccessor,spAccessible,
										a_element,a_name,
										a_message,a_create);
							}
		BWORD				access(sp<SurfaceI>& a_rspSurface,
									sp<SurfaceAccessibleI>& a_rspAccessible,
									Message a_message=e_error,
									I32 a_subIndex= -1)
							{
								a_rspSurface=a_subIndex<0?
										a_rspAccessible->surface():
										a_rspAccessible->subSurface(a_subIndex);
								if(a_rspSurface.isValid())
								{
									return TRUE;
								}
								reportSurface(a_message);
								return FALSE;
							}
		BWORD				access(sp<SurfaceI>& a_rspSurface,
									String a_key,Message a_message=e_error,
									I32 a_subIndex= -1)
							{
								sp<SurfaceAccessibleI> spAccessible;
								if(!access(spAccessible,a_key,a_message))
								{
									return FALSE;
								}

								return access(a_rspSurface,spAccessible,
										a_message,a_subIndex);
							}
		BWORD				access(sp<SurfaceAccessibleI>& a_rspAccessible,
									String a_key,Message a_message=e_error)
							{
								m_key=a_key;
								a_rspAccessible=catalog< sp<Component> >(m_key);
								if(a_rspAccessible.isValid())
								{
									return TRUE;
								}
								reportAccessible(a_message);
								return FALSE;
							}
		BWORD				access(sp<SurfaceAccessorI>& a_rspAccessor,
									sp<SurfaceAccessibleI>& a_rspAccessible,
									Element a_element,Attribute a_attribute,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								a_rspAccessor=a_rspAccessible->accessor(
										SurfaceAccessibleI::Element(a_element),
										SurfaceAccessibleI::Attribute(
										a_attribute),
										SurfaceAccessibleI::Creation(a_create));
								if(a_rspAccessor.isValid())
								{
									return TRUE;
								}
								String name;
								switch(a_attribute)
								{
									case e_generic:
										name="Generic";
										break;
									case e_position:
										name="Position";
										break;
									case e_normal:
										name="Normal";
										break;
									case e_vertices:
										name="Vertices";
										break;
									default:
										name="<bad attribute>";
										break;
								}
								reportAccessor(a_element,name,a_message);
								return FALSE;
							}
		BWORD				access(sp<SurfaceAccessorI>& a_rspAccessor,
									sp<SurfaceAccessibleI>& a_rspAccessible,
									Element a_element,String a_name,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								a_rspAccessor=a_rspAccessible->accessor(
										SurfaceAccessibleI::Element(a_element),
										a_name,
										SurfaceAccessibleI::Creation(a_create));
								if(a_rspAccessor.isValid())
								{
									return TRUE;
								}
								reportAccessor(a_element,"\""+a_name+"\"",
										a_message);
								return FALSE;
							}
		BWORD				access(sp<SurfaceAccessorI>& a_rspAccessor,
									String a_key,
									Element a_element,Attribute a_attribute,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								sp<SurfaceAccessibleI> spAccessible;
								if(!access(spAccessible,a_key,a_message))
								{
									return FALSE;
								}

								return access(a_rspAccessor,spAccessible,
										a_element,a_attribute,
										a_message,a_create);
							}
		BWORD				access(sp<SurfaceAccessorI>& a_rspAccessor,
									String a_key,
									Element a_element,String a_name,
									Message a_message=e_error,
									Creation a_create=e_createMissing)
							{
								sp<SurfaceAccessibleI> spAccessible;
								if(!access(spAccessible,a_key,a_message))
								{
									return FALSE;
								}

								return access(a_rspAccessor,spAccessible,
										a_element,a_name,
										a_message,a_create);
							}

	private:
		void				report(String a_text,Message a_message)
							{
#if FE_CODEGEN<=FE_DEBUG
								feLog("%s %s\n",name().c_str(),a_text.c_str());
#endif
								catalog<String>(a_message==e_error?
										"error": "warning")+=a_text+"; ";
							}
		void				reportDraw(String a_name,Message a_message)
							{
								if(a_message==e_quiet)
								{
									return;
								}
								String text;
								text.sPrintf("failed to access %s"
										" drawing interface",a_name.c_str());
								report(text,a_message);
							}
		void				reportSurface(Message a_message)
							{
								if(a_message==e_quiet)
								{
									return;
								}
								String text;
								text.sPrintf("failed to access SurfaceI");
								report(text,a_message);
							}
		void				reportAccessible(Message a_message)
							{
								if(a_message==e_quiet)
								{
									return;
								}
								String text;
								text.sPrintf("failed to access \"%s\"",
										m_key.c_str());
								report(text,a_message);
							}
		void				reportAccessor(Element a_element,String a_name,
									Message a_message)
							{
								if(a_message==e_quiet)
								{
									return;
								}
								String elementName;
								switch(a_element)
								{
									case e_point:
										elementName="Point";
										break;
									case e_pointGroup:
										elementName="PointGroup";
										break;
									case e_vertex:
										elementName="Vertex";
										break;
									case e_primitive:
										elementName="Primitive";
										break;
									case e_primitiveGroup:
										elementName="PrimitiveGroup";
										break;
									case e_detail:
										elementName="Detail";
										break;
									default:
										elementName="<bad element>";
										break;
								}
								String text;
								text.sPrintf("failed to access %s %s on %s",
										elementName.c_str(),
										a_name.c_str(),
										m_key.c_str()[0]=='<'? m_key.c_str():
										("\""+m_key+"\"").c_str());
								report(text,a_message);
							}

		Accessor< sp<Component> >	m_aSurfaceOutput;
		Accessor< sp<Component> >	m_aDrawOutput;
		Accessor< sp<Component> >	m_aDrawGuide;
		Accessor< sp<Component> >	m_aDrawBrush;
		Accessor<Record>			m_aWindowEvent;
		Accessor<SpatialTransform>	m_aCameraTransform;
		Accessor<SpatialVector>		m_aRayOrigin;
		Accessor<SpatialVector>		m_aRayDirection;
		String						m_key;

	private:
//		sp<DrawI>	m_spDrawI;
};

} // namespace

#endif /* __operator_OperatorSurfaceCommon_h__ */
