/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceCurve_h__
#define __surface_SurfaceCurve_h__

namespace fe
{

/**************************************************************************//**
	@brief Segmented Path Surface

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceCurves: public SurfaceSphere
{
	public:
							SurfaceCurves(void);
virtual						~SurfaceCurves(void)							{}

							//* As Protectable
virtual	Protectable*		clone(Protectable* pInstance=NULL);

							//* As SurfaceI
virtual SpatialTransform	sample(Vector2 a_uv) const;

virtual	sp<ImpactI>			nearestPoint(const SpatialVector& a_origin) const
							{
								return sp<Impact>(NULL);
							}

virtual	sp<ImpactI>			rayImpact(const SpatialVector& a_origin,
									const SpatialVector& a_direction,
									Real a_maxDistance) const
							{
								return sp<Impact>(NULL);
							}

virtual	void				draw(const SpatialTransform& a_transform,
									sp<DrawI> a_spDrawI,
									const fe::Color* a_color) const;

	protected:

	class Curve
	{
		public:
							Curve(U32 a_start, U32 a_count):
								m_start(a_start),
								m_count(a_count)							{}

			U32				start(void) const	{ return m_start; }
			U32				count(void) const	{ return m_count; }

		private:
			U32				m_start;
			U32				m_count;
	};

		U32					m_vertices;
		SpatialVector*		m_pVertexArray;
		std::vector<Curve>	m_curveArray;

virtual	void				cache(void);
};

} // namespace

#endif /* __surface_SurfaceCurve_h__ */


